@@ -1,6 +1,13 @@
 #include "python_cubrid.h"
 #include <fcntl.h>
 
+/* Loading dynamic library need this header. */
+#ifdef MS_WINDOWS
+#include <windows.h>
+#else
+#include <dlfcn.h>
+#endif
+
 #ifndef Py_TYPE
 #define Py_TYPE(ob) (((PyObject*)(ob))->ob_type)
 #endif
@@ -18,16 +25,12 @@
 #define CUBRID_ER_MSG_LEN 1024
 
 static PyObject *_cubrid_error;
-static PyObject *_cubrid_warning;
 static PyObject *_cubrid_interface_error;
 static PyObject *_cubrid_database_error;
-static PyObject *_cubrid_internal_error;
-static PyObject *_cubrid_operational_error;
-static PyObject *_cubrid_programming_error;
-static PyObject *_cubrid_integrity_error;
-static PyObject *_cubrid_data_error;
 static PyObject *_cubrid_not_supported_error;
 
+static PyObject *_func_Decimal;
+
 static struct _cubrid_isolation
 {
   int level;
@@ -106,6 +109,23 @@ static struct _error_message
   0, ""}
 };
 
+static void *libcascci = NULL;
+typedef int (*CCI_GET_LAST_INSERT_ID) (int con, void *buff,
+                                       T_CCI_ERROR * err_buf);
+static CCI_GET_LAST_INSERT_ID cci_get_last_insert_id_fp = NULL;
+
+
+#if (PY_MAJOR_VERSION == 2 && PY_MINOR_VERSION <= 4)
+typedef int Py_ssize_t;
+#endif
+
+static PyObject *
+_cubrid_return_PyUnicode_FromString (const char *buf, Py_ssize_t size, 
+        const char *encoding, const char *errors)
+{
+  return PyUnicode_Decode (buf, size, encoding, errors);
+}
+
 static PyObject *
 _cubrid_return_PyString_FromString (const char *buf)
 {
@@ -152,11 +172,13 @@ get_error_msg (int err_code, char *err_msg)
 PyObject *
 handle_error (int e, T_CCI_ERROR * error)
 {
-  PyObject *t;
+  PyObject *t, *exception = NULL;
   int err_code;
   char msg[CUBRID_ER_MSG_LEN] = { '0' };
   char err_msg[CUBRID_ER_MSG_LEN] = { '\0' }, *facility_msg;
 
+  exception = _cubrid_error;
+
   if (e == CCI_ER_DBMS)
     {
       facility_msg = "DBMS";
@@ -164,15 +186,19 @@ handle_error (int e, T_CCI_ERROR * error)
 	{
 	  err_code = error->err_code;
 	  snprintf (err_msg, CUBRID_ER_MSG_LEN, "%s", error->err_msg);
+          exception = _cubrid_database_error;
 	}
       else
 	{
 	  err_code = 0;
 	  snprintf (err_msg, CUBRID_ER_MSG_LEN, "Unknown DBMS Error");
+          exception = _cubrid_not_supported_error;
 	}
     }
   else
     {
+      exception = _cubrid_interface_error;
+
       if (get_error_msg (e, err_msg) < 0)
 	{
 	  snprintf (err_msg, CUBRID_ER_MSG_LEN, "Unknown Error");
@@ -203,10 +229,10 @@ handle_error (int e, T_CCI_ERROR * error)
   if (!(t = PyTuple_New (2)))
     return NULL;
 
-  PyTuple_SetItem (t, 0, _cubrid_return_PyInt_FromLong ((long) e));
+  PyTuple_SetItem (t, 0, _cubrid_return_PyInt_FromLong ((long) err_code));
   PyTuple_SetItem (t, 1, _cubrid_return_PyString_FromString (msg));
 
-  PyErr_SetObject (_cubrid_error, t);
+  PyErr_SetObject (exception, t);
   Py_DECREF (t);
 
   return NULL;
@@ -261,82 +287,6 @@ static char _cubrid_escape_string__doc__[] =
 "escape_string()\n\
 Escape special characters in a string for use in an SQL statement";
 
-#ifdef _WIN32
-static PyObject *
-_cubrid_escape_string (PyObject * self, PyObject *args,
-                       PyObject * kwargs)
-{
-  static char *kwList[] = {"escape_string", "no_backslash_escapes", NULL};
-  int no_backslash_escapes = -1;
-  char *unescape_string = NULL, *escape_string = NULL;
-  int len = -1, res;
-  T_CCI_ERROR error;
-  
-    PyObject *op;
-
-    int escape_str_len = 0, i = 0;
-    char *s = NULL;
-
-
-  if (!PyArg_ParseTupleAndKeywords (args, kwargs,
-            "s#|i", kwList, &unescape_string, &len, &no_backslash_escapes))
-    {
-      return NULL;
-    }
-	
-    if (len < 0)
-    {
-        return handle_error (CUBRID_ER_INVALID_PARAM, NULL);
-    }
-
-    for (i = 0; i < len; i++) {
-        if (unescape_string[i] == '\'') {
-            escape_str_len++;
-        }
-        escape_str_len++;
-    }
-
-    if (escape_str_len == len) {
-#if PY_MAJOR_VERSION >= 3
-        op = PyUnicode_FromStringAndSize (unescape_string, len);
-#else
-        op = PyString_FromStringAndSize (unescape_string, len);
-#endif
-        return op;
-    }
-
-    escape_string = (char *) malloc (len * 2 + 16);
-
-    if (!escape_string)
-    {
-        return handle_error (CUBRID_ER_NO_MORE_MEMORY, NULL);
-    }
-
-    memset(escape_string, 0, len * 2 + 16);
-
-    s = escape_string;
-    for (i = 0; i < len; i++) {
-        if (unescape_string[i] == '\'') {
-            *s++ = '\'';
-            *s++ = '\'';
-        } else {
-            *s++ = unescape_string[i];
-        }
-    }
-
-    escape_string[escape_str_len] = '\0';
-
-#if PY_MAJOR_VERSION >= 3
-    op = PyUnicode_FromStringAndSize (escape_string, escape_str_len);
-#else
-    op = PyString_FromStringAndSize (escape_string, escape_str_len);
-#endif
-
-    free (escape_string);
-
-    return op;
-}
-#else
 static PyObject *
 _cubrid_escape_string (PyObject * self, PyObject *args,
                        PyObject * kwargs)
@@ -394,7 +344,6 @@ _cubrid_escape_string (PyObject * self, PyObject *args,
 
   return op;
 }
-#endif
 
 static PyObject *
 _cubrid_ConnectionObject_new (PyTypeObject * type, PyObject * args,
@@ -436,21 +385,12 @@ _cubrid_ConnectionObject_init (_cubrid_ConnectionObject * self,
 
   snprintf (buf, 1024, "cci:%s", url);
 
-#ifdef _WIN32
-  con = cci_connect_with_url (buf, user, passwd);
-  if (con < 0)
-    {
-      handle_error (con, NULL);
-      return -1;
-    }
-#else
   con = cci_connect_with_url_ex (buf, user, passwd, &error);
   if (con < 0)
     {
       handle_error (con, &error);
       return -1;
     }
-#endif
 
   self->handle = con;
 
@@ -521,7 +461,7 @@ Get the cursor class. Return a new Cursor Object.\n\
 \n\
 Example::\n\
   import _cubrid\n\
-  con = _cubrid.connect(\"CUBRID:localhost:33000:demodb\", \"public\")\n\
+  con = _cubrid.connect(\"CUBRID:localhost:33000:demodb:::\", \"public\")\n\
   cur = con.cursor()\n\
   ...\n\
   other operations\n\
@@ -564,7 +504,7 @@ Create a large object. Return a new lob object.\n\
 \n\
 Example::\n\
   import _cubrid\n\
-  con = _cubrid.connect(\"CUBRID:localhost:33000:demodb\", \"public\")\n\
+  con = _cubrid.connect(\"CUBRID:localhost:33000:demodb:::\", \"public\")\n\
   cur = con.cursor()\n\
   cur.prepare('insert into test_lob(image) values (?)')\n\
   lob = con.lob()\n\
@@ -662,7 +602,7 @@ Returns a string that represents the server version number.\n\
 \n\
 Example::\n\
   import _cubrid\n\
-  con = _cubrid.connect(\"CUBRID:localhost:33000:demodb\", \"public\")\n\
+  con = _cubrid.connect(\"CUBRID:localhost:33000:demodb:::\", \"public\")\n\
   print con.server_version()\n\
   con.close()";
 
@@ -695,7 +635,7 @@ Return a string that represents the CUBRID client library\n\
 \n\
 Example::\n\
   import _cubrid\n\
-  con = _cubrid.connect(\"CUBRID:localhost:33000:demodb\", \"public\")\n\
+  con = _cubrid.connect(\"CUBRID:localhost:33000:demodb:::\", \"public\")\n\
   print con.client_version()\n\
   con.close()";
 
@@ -727,7 +667,7 @@ mode: string. It will be ON/OFF, TRUE/FALSE, YES/NO\n\
 \n\
 Example::\n\
   import _cubrid\n\
-  con = _cubrid.connect(\"CUBRID:localhost:33000:demodb\", \"public\")\n\
+  con = _cubrid.connect(\"CUBRID:localhost:33000:demodb:::\", \"public\")\n\
   con.set_autocommit('ON')\n\
   print con.autocommit\n\
   con.close()";
@@ -795,7 +735,7 @@ isolation_level maybe::\n\
 Example::\n\
   import _cubrid\n\
   form _cubrid import *\n\
-  con = _cubrid.connect(\"CUBRID:localhost:33000:demodb\", \"public\")\n\
+  con = _cubrid.connect(\"CUBRID:localhost:33000:demodb:::\", \"public\")\n\
   con.set_isolation_level(CUBRID_REP_CLASS_REP_INSTANCE)\n\
   print con.isolation_level\n\
   con.close()";
@@ -839,7 +779,7 @@ Return values::\n\
 \n\
 Example::\n\
   import _cubrid\n\
-  con = _cubrid.connect(\"CUBRID:localhost:33000:demodb\", \"public\")\n\
+  con = _cubrid.connect(\"CUBRID:localhost:33000:demodb:::\", \"public\")\n\
   print con.ping()\n\
   con.close()";
 
@@ -918,7 +858,7 @@ columns that were updated by the previous INSERT query.\n\
 \n\
 Example::\n\
   import _cubrid\n\
-  con = _cubrid.connect(\"CUBRID:localhost:33000:demodb\", \"public\")\n\
+  con = _cubrid.connect(\"CUBRID:localhost:33000:demodb:::\", \"public\")\n\
   cur = con.curosr()\n\
   cur.prepare(\"create table test_cubrid(id NUMERIC\n\
           AUTO_INCREMENT(10300, 1), name VARCHAR(50))\")\n\
@@ -938,7 +878,17 @@ _cubrid_ConnectionObject_last_insert_id (_cubrid_ConnectionObject * self,
   int res;
   T_CCI_ERROR error;
 
-  res = cci_last_insert_id (self->handle, &name, &error);
+  if (cci_get_last_insert_id_fp != NULL) 
+    {
+      /* cci_get_last_id set last_id as con_handle's static buffer */
+      res = cci_get_last_insert_id_fp (self->handle, &name, &error);
+    }
+  else
+    {
+      /* cci_last_id set last_id as allocated string */
+      res = cci_last_insert_id (self->handle, &name, &error);
+    }
+
   if (res < 0)
     {
       return handle_error (res, &error);
@@ -952,7 +902,12 @@ _cubrid_ConnectionObject_last_insert_id (_cubrid_ConnectionObject * self,
   else
     {
       strncpy (ret, name, sizeof (ret) - 1);
-      free (name);
+      if (cci_get_last_insert_id_fp == NULL) 
+        {
+#ifndef MS_WINDOWS
+          free (name);
+#endif
+        }
     }
 
   return _cubrid_return_PyString_FromString (ret);
@@ -1009,6 +964,7 @@ _cubrid_ConnectionObject_schema_to_pyvalue (_cubrid_ConnectionObject * self,
 #endif
 	  Py_DECREF (tmpval);
 	}
+      break;
     default:
       res = cci_get_data (request, index, CCI_A_TYPE_STR, &buffer, &ind);
       if (res < 0)
@@ -1164,8 +1120,8 @@ Return values::\n\
 \n\
 Example::\n\
   import _cubrid\n\
-  con = _cubrid.connect('CUBRID:localhost:33000:demodb', 'public')\n\
-  print con.schema_info(CUBRID_SCH_TABLE, 'test_cubrid')\n\
+  con = _cubrid.connect('CUBRID:localhost:33000:demodb:::', 'public')\n\
+  print con.schema_info(_cubrid.CUBRID_SCH_TABLE, 'test_cubrid')\n\
   con.close()";
 
 static PyObject *
@@ -1258,77 +1214,6 @@ static char _cubrid_ConnectionObject_escape_string__doc__[] =
 "escape_string()\n\
 Escape special characters in a string for use in an SQL statement";
 
-#ifdef _WIN32
-static PyObject *
-_cubrid_ConnectionObject_escape_string (_cubrid_ConnectionObject * self,
-				        PyObject * args)
-{
-    char *unescape_string = NULL, *escape_string = NULL;
-    int len = -1, res;
-    PyObject *op;
-    //T_CCI_ERROR error;
-
-    int escape_str_len = 0, i = 0;
-    char *s = NULL;
-
-    if (!PyArg_ParseTuple (args, "s#", &unescape_string, &len))
-    {
-        return NULL;
-    }
-
-    if (len < 0)
-    {
-        return handle_error (CUBRID_ER_INVALID_PARAM, NULL);
-    }
-
-    for (i = 0; i < len; i++) {
-        if (unescape_string[i] == '\'') {
-            escape_str_len++;
-        }
-        escape_str_len++;
-    }
-
-    if (escape_str_len == len) {
-#if PY_MAJOR_VERSION >= 3
-        op = PyUnicode_FromStringAndSize (unescape_string, len);
-#else
-        op = PyString_FromStringAndSize (unescape_string, len);
-#endif
-        return op;
-    }
-
-    escape_string = (char *) malloc (len * 2 + 16);
-
-    if (!escape_string)
-    {
-        return handle_error (CUBRID_ER_NO_MORE_MEMORY, NULL);
-    }
-
-    memset(escape_string, 0, len * 2 + 16);
-
-    s = escape_string;
-    for (i = 0; i < len; i++) {
-        if (unescape_string[i] == '\'') {
-            *s++ = '\'';
-            *s++ = '\'';
-        } else {
-            *s++ = unescape_string[i];
-        }
-    }
-
-    escape_string[escape_str_len] = '\0';
-
-#if PY_MAJOR_VERSION >= 3
-    op = PyUnicode_FromStringAndSize (escape_string, escape_str_len);
-#else
-    op = PyString_FromStringAndSize (escape_string, escape_str_len);
-#endif
-
-    free (escape_string);
-
-    return op;
-}
-#else
 static PyObject *
 _cubrid_ConnectionObject_escape_string (_cubrid_ConnectionObject * self,
 				        PyObject * args)
@@ -1374,7 +1259,6 @@ _cubrid_ConnectionObject_escape_string (_cubrid_ConnectionObject * self,
 
   return op;
 }
-#endif
 
 static char _cubrid_ConnectionObject_close__doc__[] =
 "close()\n\
@@ -1469,9 +1353,35 @@ _cubrid_CursorObject_init (_cubrid_CursorObject * self, PyObject * args,
   self->row_count = -1;
   self->cursor_pos = 0;
 
+  memset(self->charset, 0, sizeof(self->charset));
+
   return 0;
 }
 
+static char _cubrid_CursorObject__set_charset_name__doc__[] =
+"Only used internally. This function should not be used by user.";
+
+static PyObject *
+_cubrid_CursorObject__set_charset_name (_cubrid_CursorObject * self, PyObject * args)
+{
+  char *charset = NULL;
+
+  if (!PyArg_ParseTuple (args, "s", &charset))
+    {
+      return NULL;
+    }
+
+  if (charset != NULL && *charset != '\0')
+    {
+      snprintf(self->charset, sizeof(self->charset), "%s", charset);
+    }
+
+  Py_INCREF (Py_None);
+  return Py_None;
+
+}
+
+
 static void
 _cubrid_CursorObject_reset (_cubrid_CursorObject * self)
 {
@@ -1495,12 +1405,14 @@ _cubrid_CursorObject_reset (_cubrid_CursorObject * self)
 
 static char _cubrid_CursorObject_prepare__doc__[] =
 "prepare(sql)\n\
-This function is sort of API which represents SQL statements compiled\n\
-previously. Accordingly, you can use this statement effectively to\n\
-execute several times repeatedly or to process long data Only a single\n\
-statement can be used and a parameter may put a question mark (?) to\n\
-appropriate area in the SQL statement. Add a parameter when you bind\n\
-a value in the VALUES clause of INSERT statement or in the WHERE clause.\n\
+This function creates a prepared statement. A prepared statement is a\n\
+server-side object that can be used to optimize performance. You can \n\
+use this statement effectively to execute repeatedly or to process \n\
+long data. Only a single statement can be used. The SQL statement can \n\
+contain zero or more question mark (?) parameter markers for which real \n\
+values will be substituted when the statement is executed. Add a \n\
+parameter when you bind a value in the VALUES clause of INSERT statement \n\
+or in the WHERE clause.\n\
 \n\
 sql: string, the sql statement you want to execute.";
 
@@ -1567,7 +1479,7 @@ _cubrid_CursorObject_bind_param (_cubrid_CursorObject * self, PyObject * args)
       return handle_error (CUBRID_ER_SQL_UNPREPARE, NULL);
     }
 
-  if (!PyArg_ParseTuple (args, "is", &index, &value))
+  if (!PyArg_ParseTuple (args, "iz", &index, &value))
     {
       return NULL;
     }
@@ -1600,7 +1512,7 @@ Parameters::\n\
 \n\
 Example::\n\
   import _cubrid\n\
-  con = _cubrid.connect('CUBRID:localhost:33000:demodb', 'public')\n\
+  con = _cubrid.connect('CUBRID:localhost:33000:demodb:::', 'public')\n\
   cur = con.cursor()\n\
   \n\
   cur.prepare('create table test_blob(image BLOB)')\n\
@@ -1754,7 +1666,7 @@ row: int, the column you want to get the information.\n\
 \n\
 Example::\n\
   import _cubrid\n\
-  con = _cubrid.connect('CUBRID:localhost:33000:demodb', 'public')\n\
+  con = _cubrid.connect('CUBRID:localhost:33000:demodb:::', 'public')\n\
   cur = con.cursor()\n\
   cur.prepare('select * from test_cubrid')\n\
   cur.execute()\n\
@@ -1850,15 +1762,14 @@ _cubrid_CursorObject_result_info (_cubrid_CursorObject * self,
 
 static char _cubrid_CursorObject_execute__doc__[] =
 "execute([option[,max_col_size]])\n\
-execute the prepared SQL statement, which is executing prepare() and\n\
-bind_param()/bind_lob(). The function of retrieving the query result\n\
-from the server through a flag can be classified as synchronous or\n\
-asynchronous. If the flag is set to CUBRID_EXEC_QUERY_ALL, a\n\
-synchronous mode(sync_mode) is used to retrieve query results\n\
-immediately after executing prepared queries. If it is set to\n\
-CUBRID_EXEC_ASYNC, an asynchronous mode (async_mode) is used to\n\
+Executes a prepared Query.\n\
+A option can be used when retrieving the query result from the server.\n\
+A option can be classified as synchronous or asynchronous. \n\
+If the option is set to CUBRID_EXEC_QUERY_ALL, a synchronous mode(sync_mode) \n\
+is used to retrieve query results immediately after executing prepared queries. \n\
+If it is set to CUBRID_EXEC_ASYNC, an asynchronous mode (async_mode) is used to\n\
 retrieve the result immediately each time a query result is created.\n\
-The flag is set to CUBRID_EXEC_QUERY_ALL by default, and in such\n\
+The option is set to CUBRID_EXEC_QUERY_ALL by default, and in such\n\
 cases the following rules are applied:\n\
   - The return value is the result of the first query.\n\
   - If an error occurs in any query, the execution is processed\n\
@@ -1875,7 +1786,7 @@ prepared query is CHAR, VARCHAR, NCHAR, VARNCHAR, BIT or VARBIT.\n\
 If it is set to 0, all data is transferred.\n\
 \n\
 Parameters::\n\
-  flag: Exec flag, flag maybe the following values:\n\
+  option: Exec option, option maybe the following values:\n\
     CUBRID_EXEC_ASYNC\n\
     CUBRID_EXEC_QUERY_ALL\n\
     CUBRID_EXEC_QUERY_INFO\n\
@@ -1935,11 +1846,13 @@ _cubrid_CursorObject_execute (_cubrid_CursorObject * self, PyObject * args)
 
   if (res_sql_type == SQLX_CMD_SELECT)
     {
+      int ret;
+
       _cubrid_CursorObject_set_description (self);
-      res = cci_cursor (self->handle, 1, CCI_CURSOR_CURRENT, &error);
-      if (res < 0 && res != CCI_ER_NO_MORE_DATA)
+      ret = cci_cursor (self->handle, 1, CCI_CURSOR_CURRENT, &error);
+      if (ret < 0 && ret != CCI_ER_NO_MORE_DATA)
 	{
-	  return handle_error (res, &error);
+	  return handle_error (ret, &error);
 	}
     }
 
@@ -1950,6 +1863,11 @@ _cubrid_CursorObject_execute (_cubrid_CursorObject * self, PyObject * args)
 * 
 * int, short 			-> Integer
 * float, double, numeric 	-> Float
+* numeric   			-> Decimal
+* time 					-> datetime.time
+* date 					-> datetime.date
+* datetime 				-> datetime.datetime
+* timestamp 			-> datetime.datetime
 * another type			-> String
 */
 
@@ -1961,6 +1879,7 @@ _cubrid_CursorObject_dbval_to_pyvalue (_cubrid_CursorObject * self, int type,
   PyObject *val, *tmpval;
   char *buffer;
   int num;
+  T_CCI_DATE dt;
 
   switch (type)
     {
@@ -1983,7 +1902,6 @@ _cubrid_CursorObject_dbval_to_pyvalue (_cubrid_CursorObject * self, int type,
       break;
     case CCI_U_TYPE_FLOAT:
     case CCI_U_TYPE_DOUBLE:
-    case CCI_U_TYPE_NUMERIC:
       res = cci_get_data (self->handle, index, CCI_A_TYPE_STR, &buffer, &ind);
       if (res < 0)
 	{
@@ -2004,7 +1922,90 @@ _cubrid_CursorObject_dbval_to_pyvalue (_cubrid_CursorObject * self, int type,
 #endif
 	  Py_DECREF (tmpval);
 	}
-      break;
+	  break;
+    case CCI_U_TYPE_NUMERIC:
+	  res = cci_get_data (self->handle, index, CCI_A_TYPE_STR, &buffer, &ind);
+	  if (res < 0) 
+	{
+	   return handle_error (res, NULL);
+	}
+	  if (ind < 0) 
+	{
+	  Py_INCREF (Py_None);
+	  val = Py_None;
+	}
+	  else
+	{
+	  tmpval = PyTuple_New (1);
+	  PyTuple_SetItem (tmpval, 0, Py_BuildValue ("s", buffer));
+	  val = PyObject_CallObject(_func_Decimal, tmpval);
+	  Py_DECREF (tmpval);
+	}
+	  break;
+    case CCI_U_TYPE_DATE:
+	  res = cci_get_data (self->handle, index, CCI_A_TYPE_DATE, &dt, &ind);
+	  if (res < 0)
+	{
+	  return handle_error (res, NULL);
+	}
+	  if (ind < 0)
+	{
+	  Py_INCREF (Py_None);
+	  val = Py_None;
+	}
+	  else
+	{
+	  val = PyDate_FromDate (dt.yr, dt.mon, dt.day);
+	}
+	  break;
+    case CCI_U_TYPE_TIME:
+	  res = cci_get_data (self->handle, index, CCI_A_TYPE_DATE, &dt, &ind);
+	  if (res < 0)
+	{
+	  return handle_error (res, NULL);
+	}
+	  if (ind < 0)
+	{
+	  Py_INCREF (Py_None);
+	  val = Py_None;
+	}
+	  else
+	{
+	  val = PyTime_FromTime (dt.hh, dt.mm, dt.ss, 0);
+	}
+	  break;
+    case CCI_U_TYPE_DATETIME:
+	  res = cci_get_data (self->handle, index, CCI_A_TYPE_DATE, &dt, &ind);
+	  if (res < 0)
+	{
+	  return handle_error (res, NULL);
+	}
+	  if (ind < 0)
+	{
+	  Py_INCREF (Py_None);
+	  val = Py_None;
+	}
+	  else
+	{
+	  val = PyDateTime_FromDateAndTime (dt.yr, dt.mon, dt.day, dt.hh, dt.mm, dt.ss, dt.ms * 1000);
+	}
+	  break;
+    case CCI_U_TYPE_TIMESTAMP:
+	  res = cci_get_data (self->handle, index, CCI_A_TYPE_DATE, &dt, &ind);
+	  if (res < 0)
+	{
+	  return handle_error (res, NULL);
+	}
+	  if (ind < 0)
+	{
+	  Py_INCREF (Py_None);
+	  val = Py_None;
+	}
+	  else
+	{
+	  val = PyDateTime_FromDateAndTime (dt.yr, dt.mon, dt.day, dt.hh, dt.mm, dt.ss, 0);
+	}
+	  break;
     default:
       res = cci_get_data (self->handle, index, CCI_A_TYPE_STR, &buffer, &ind);
       if (res < 0)
@@ -2018,7 +2019,14 @@ _cubrid_CursorObject_dbval_to_pyvalue (_cubrid_CursorObject * self, int type,
 	}
       else
 	{
-	  val = _cubrid_return_PyString_FromString (buffer);
+          if (self->charset != NULL && *(self->charset) != '\0')
+            {
+              val = _cubrid_return_PyUnicode_FromString (buffer, strlen(buffer), self->charset, NULL);
+            }
+          else
+            {
+              val = _cubrid_return_PyString_FromString (buffer);
+            }
 	}
       break;
     }
@@ -2135,7 +2143,7 @@ to the next row after getting the result.\n\
 \n\
 Example::\n\
   import _cubrid\n\
-  con = _cubrid.connect('CUBRID:localhost:33000:demodb', 'public')\n\
+  con = _cubrid.connect('CUBRID:localhost:33000:demodb:::', 'public')\n\
   cur = con.cursor()\n\
   cur.prepare('select * from test_cubrid')\n\
   cur.execute()\n\
@@ -2211,7 +2219,7 @@ Parameters::\n\
 \n\
 Example::\n\
   import _cubrid\n\
-  con = _cubrid.connect('CUBRID:localhost:33000:demodb', 'public')\n\
+  con = _cubrid.connect('CUBRID:localhost:33000:demodb:::', 'public')\n\
   cur = con.cursor()\n\
   cur.prepare('select * from test_lob')\n\
   cur.execute()\n\
@@ -2736,7 +2744,7 @@ It will be create a BLOB object as default.\n\
 \n\
 Example 1::\n\
   import _cubrid\n\
-  con = _cubrid.connect('CUBRID:localhost:33000:demodb', 'public')\n\
+  con = _cubrid.connect('CUBRID:localhost:33000:demodb:::', 'public')\n\
   cur = con.cursor()\n\
   cur.prepare('insert into test_clob(content) values (?)')\n\
   lob = con.lob()\n\
@@ -2750,7 +2758,7 @@ Example 1::\n\
 \n\
 Example 2::\n\
   import _cubrid\n\
-  con = _cubrid.connect('CUBRID:localhost:33000:demodb', 'public')\n\
+  con = _cubrid.connect('CUBRID:localhost:33000:demodb:::', 'public')\n\
   cur = con.cursor()\n\
   cur.prepare('select * from test_blob')\n\
   cur.execute()\n\
@@ -2822,7 +2830,7 @@ file: string, support filepath/file\n\
 \n\
 Example::\n\
   import _cubrid\n\
-  con = _cubrid.connect('CUBRID:localhost:33000:demodb', 'public')\n\
+  con = _cubrid.connect('CUBRID:localhost:33000:demodb:::', 'public')\n\
   cur = con.cursor()\n\
   cur.prepare('select * from test_lob')\n\
   cur.execute()\n\
@@ -2898,7 +2906,7 @@ Return a string that contains the data read.\n\
 \n\
 Example 1::\n\
   import _cubrid\n\
-  con = _cubrid.connect('CUBRID:localhost:33000:demodb', 'public')\n\
+  con = _cubrid.connect('CUBRID:localhost:33000:demodb:::', 'public')\n\
   lob = con.lob()\n\
   lob.imports('README', 'C')\n\
   str = lob.read(32)\n\
@@ -2908,7 +2916,7 @@ Example 1::\n\
 \n\
 Example 2::\n\
   import _cubrid\n\
-  con = _cubrid.connect('CUBRID:localhost:33000:demodb', 'public')\n\
+  con = _cubrid.connect('CUBRID:localhost:33000:demodb:::', 'public')\n\
   cur = con.cursor()\n\
   cur.prepare('select * from test_lob')\n\
   cur.execute()\n\
@@ -3126,6 +3134,11 @@ static PyMethodDef _cubrid_CursorObject_methods[] = {
    (PyCFunction) _cubrid_CursorObject_prepare,
    METH_VARARGS,
    _cubrid_CursorObject_prepare__doc__},
+  {
+   "_set_charset_name",
+   (PyCFunction) _cubrid_CursorObject__set_charset_name,
+   METH_VARARGS,
+   _cubrid_CursorObject__set_charset_name__doc__},
   {
    "bind_param",
    (PyCFunction) _cubrid_CursorObject_bind_param,
@@ -3444,7 +3457,7 @@ PyTypeObject _cubrid_CursorObject_type = {
   0,				/* tp_free */
 };
 
-#define _CUBRID_VERSION_	"9.0.0.0001"
+#define _CUBRID_VERSION_	"9.1.0.0001"
 static char _cubrid_doc[] = "CUBRID API Module for Python";
 
 #if PY_MAJOR_VERSION >= 3
@@ -3468,10 +3481,6 @@ init_exceptions (PyObject * dict)
     PyErr_NewException ("_cubrid.Error", PyExc_Exception, NULL);
   PyDict_SetItemString (dict, "Error", _cubrid_error);
 
-  _cubrid_warning =
-    PyErr_NewException ("_cubrid.Warning", PyExc_Exception, NULL);
-  PyDict_SetItemString (dict, "Warning", _cubrid_warning);
-
   _cubrid_interface_error =
     PyErr_NewException ("_cubrid.InterfaceError", _cubrid_error, NULL);
   PyDict_SetItemString (dict, "InterfaceError", _cubrid_interface_error);
@@ -3480,35 +3489,9 @@ init_exceptions (PyObject * dict)
     PyErr_NewException ("_cubrid.DatabaseError", _cubrid_error, NULL);
   PyDict_SetItemString (dict, "DatabaseError", _cubrid_database_error);
 
-  _cubrid_internal_error =
-    PyErr_NewException ("_cubrid.InternalError", _cubrid_database_error,
-			NULL);
-  PyDict_SetItemString (dict, "InternalError", _cubrid_internal_error);
-
-  _cubrid_operational_error =
-    PyErr_NewException ("_cubrid.OperationalError", _cubrid_database_error,
-			NULL);
-  PyDict_SetItemString (dict, "OperationalError", _cubrid_operational_error);
-
-  _cubrid_programming_error =
-    PyErr_NewException ("_cubrid.ProgrammingError", _cubrid_database_error,
-			NULL);
-  PyDict_SetItemString (dict, "ProgrammingError", _cubrid_programming_error);
-
-  _cubrid_integrity_error =
-    PyErr_NewException ("_cubrid.IntegrityError", _cubrid_database_error,
-			NULL);
-  PyDict_SetItemString (dict, "IntegrityError", _cubrid_integrity_error);
-
-  _cubrid_data_error =
-    PyErr_NewException ("_cubrid.DataError", _cubrid_database_error, NULL);
-  PyDict_SetItemString (dict, "DataError", _cubrid_data_error);
-
   _cubrid_not_supported_error =
-    PyErr_NewException ("_cubrid.NotSupportedError", _cubrid_database_error,
-			NULL);
-  PyDict_SetItemString (dict, "NotSupportedError",
-			_cubrid_not_supported_error);
+    PyErr_NewException ("_cubrid.NotSupportedError", _cubrid_database_error,NULL);
+  PyDict_SetItemString (dict, "NotSupportedError", _cubrid_not_supported_error);
 }
 
 static int
@@ -3649,7 +3632,30 @@ void
 init_cubrid (void)
 #endif
 {
-  PyObject *dict, *module;
+  PyObject *dict, *module, *mDecimal;
+
+  /* 
+   * use cci_get_last_id if possible
+   * early distributed libraries don't include it
+   */
+#ifdef MS_WINDOWS
+    libcascci = LoadLibrary ("cascci.dll");
+
+    if (libcascci != NULL) 
+      {
+        cci_get_last_insert_id_fp =
+            (CCI_GET_LAST_INSERT_ID) GetProcAddress (libcascci, 
+                                                    "cci_get_last_insert_id");
+      }
+#else
+    libcascci = dlopen ("libcascci.so", RTLD_LAZY);
+
+    if (libcascci != NULL) 
+      {
+        cci_get_last_insert_id_fp = dlsym (libcascci, "cci_get_last_insert_id");
+      }
+#endif
+
 
 #if PY_MAJOR_VERSION >= 3
   module = PyModule_Create (&cubriddef);
@@ -3706,6 +3712,25 @@ init_cubrid (void)
       goto Error;
     }
 
+ /* import function Decimal from module decimal */
+  mDecimal = PyImport_ImportModule ("decimal");
+  if (!mDecimal)
+    {
+      goto Error;
+    }
+
+  _func_Decimal = PyObject_GetAttrString (mDecimal, "Decimal");
+  if (!_func_Decimal)
+    {
+      goto Error;
+    }
+
+  Py_INCREF (_func_Decimal);
+  Py_DECREF (mDecimal);
+
+  /* invoke PyDateTime_IMPORT macro to use functions from datetime.h */
+  PyDateTime_IMPORT;
+
 #if PY_MAJOR_VERSION >= 3
   return module;
 #else
